from dataclasses import dataclass
from typing import Any, Callable, Optional
from enum import Enum
import operator


@dataclass
class BinOperator:
    symbol: str
    priority: int
    perform: Callable[[float, float], float]


OPERATORS = [
    BinOperator("+", 0, operator.add),
    BinOperator("-", 0, operator.sub),
    BinOperator("*", 1, operator.mul),
    BinOperator("/", 1, operator.truediv)
]

# ops_by_priority = [list(it) for _, it in groupby(OPERATORS, lambda op: op.priority)]
ops_by_priority = [
    [OPERATORS[0], OPERATORS[1]],
    [OPERATORS[2], OPERATORS[3]]
]
MAX_PRIORITY = len(ops_by_priority)
ops_syms = [op.symbol for op in OPERATORS]


class TokenType(Enum):
    NUMBER = 1
    PARENTHESIS = 2
    OPERATION = 3


@dataclass
class Token:
    type: TokenType
    val: str
    num: float


def tokenize(inp: str):
    tokens = []

    index = 0

    def skip_spaces():
        nonlocal index
        while inp[index].isspace():
            index += 1

    def has():
        return index < len(inp)

    def peek():
        return inp[index]

    def read():
        nonlocal index
        index += 1
        return inp[index - 1]

    def read_number():
        res = ""

        while True:
            res += read()
            if not has() or peek() not in "0123456789.":
                break

        return Token(TokenType.NUMBER, res, float(res))  # if "." in res else int(res))

    while has():
        skip_spaces()

        next = peek()

        tok: Token

        if next in ops_syms:
            tok = Token(TokenType.OPERATION, read(), 0.)
        elif next in "()":
            tok = Token(TokenType.PARENTHESIS, read(), 0.)
        elif next in "0123456789.":
            tok = read_number()
        else:
            raise Exception("invalid character '{}' at {}".format(next, index))

        tokens.append(tok)

    return tokens


class Parser:
    tokens: list[Token]
    index: int

    def __init__(self, tokens):
        self.tokens = tokens
        self.index = 0

    def has(self):
        return self.index < len(self.tokens)

    def current(self):
        if not self.has():
            raise Exception("expected token, got EOL")
        return self.tokens[self.index]

    def match(self, type: TokenType, val: Optional[str] = None):
        return self.has() and self.tokens[self.index].type == type and (val is None or self.tokens[self.index].val == val)

    def accept(self, type: TokenType, val: Optional[str] = None):
        if self.match(type, val):
            self.index += 1
            return True
        return False

    def expect(self, type: TokenType, val: Optional[str] = None):
        if self.match(type, val):
            self.index += 1
            return self.tokens[self.index - 1]
        if not self.has():
            raise Exception("expected {}, got EOL".format(type))
        else:
            raise Exception("expected {}, got {}".format(type, self.current().type))

    def parse_bin(self, priority=0):
        if priority >= MAX_PRIORITY:
            return self.parse_term()

        left = self.parse_bin(priority + 1)
        ops = ops_by_priority[priority]

        while self.has() and self.current().type == TokenType.OPERATION:
            for op in ops:
                if self.accept(TokenType.OPERATION, op.symbol):
                    right = self.parse_bin(priority + 1)
                    left = op.perform(left, right)
                    break
            else:
                break

        return left

    def parse_expr(self):
        return self.parse_bin()

    def parse_term(self):
        token = self.current()

        if token.type == TokenType.NUMBER:
            return self.expect(TokenType.NUMBER).num
        elif self.accept(TokenType.PARENTHESIS, "("):
            val = self.parse_expr()
            self.expect(TokenType.PARENTHESIS, ")")
            return val
        else:
            raise Exception("expected term, got {}".format(token.type))

def parse(tokens):
    return Parser(tokens).parse_expr()

if __name__ == "__main__":
    while True:
        # inp = input("> ")
        inp = "2 + 3 * 4"
        try:
            tok = tokenize(inp)
            res = parse(tok)
            print(res)
        except Exception as e:
            # print(e)
            pass
        print()
        break
