import dataclasses
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from itertools import zip_longest
from typing import Dict, Optional, List, ClassVar, Callable
from transpiler.utils import highlight


def get_default_parents():
    if obj := globals().get("TY_OBJECT"):
        return [obj]
    return []


class RuntimeValue:
    pass

@dataclass
class MemberDef:
    type: "BaseType"
    val: typing.Any = RuntimeValue()
    in_class_def: bool = True

@dataclass
class UnifyMode:
    search_hierarchy: bool = True
    match_protocol: bool = True


UnifyMode.NORMAL = UnifyMode()
UnifyMode.EXACT = UnifyMode(False, False)

@dataclass(eq=False)
class BaseType(ABC):
    #members: Dict[str, "MemberDef"] = field(default_factory=dict, init=False)
    #methods: Dict[str, "FunctionType"] = field(default_factory=dict, init=False)
    fields: Dict[str, "MemberDef"] = field(default_factory=dict, init=False)
    parents: List["BaseType"] = field(default_factory=get_default_parents, init=False)
    typevars: List["TypeVariable"] = field(default_factory=list, init=False)
    #static_members: Dict[str, "MemberDef"] = field(default_factory=dict, init=False)

    def get_members(self):
        return {n: m for n, m in self.fields.items() if type(m.val) is RuntimeValue}


    def get_parents(self) -> List["BaseType"]:
        return self.parents

    def iter_hierarchy_recursive(self) -> typing.Iterator["BaseType"]:
        cache = set()
        from queue import Queue
        queue = Queue()
        queue.put(self)
        while not queue.empty():
            cur = queue.get()
            yield cur
            if cur in cache:
                continue
            cache.add(cur)
            if cur == TY_OBJECT:
                continue
            for p in cur.get_parents():
                queue.put(p)

    def inherits_from(self, other: "BaseType") -> bool:
        from transpiler.exceptions import CompileError
        for parent in self.iter_hierarchy_recursive():
            try:
                parent.unify(other, UnifyMode.EXACT)
            except CompileError:
                pass
            else:
                return True
        return False

    def resolve(self) -> "BaseType":
        return self

    @abstractmethod
    def unify_internal(self, other: "BaseType", mode: UnifyMode):
        pass

    def unify(self, other: "BaseType", mode = UnifyMode.NORMAL):
        a, b = self.resolve(), other.resolve()
        TB = f"unifying {highlight(a)} and {highlight(b)}"
        if isinstance(b, TypeVariable):
            a, b = b, a
        a.unify_internal(b, mode)

    def contains(self, other: "BaseType") -> bool:
        needle, haystack = other.resolve(), self.resolve()
        return (needle is haystack) or haystack.contains_internal(needle)

    @abstractmethod
    def contains_internal(self, other: "BaseType") -> bool:
        pass

    # @abstractmethod
    # def clone(self) -> "BaseType":
    #     pass

    def gen_sub(self, this: "BaseType", typevars: Dict[str, "BaseType"], cache=None) -> "Self":
        return self

    def to_list(self) -> List["BaseType"]:
        return [self]


T = typing.TypeVar("T")


class MagicType(BaseType, typing.Generic[T]):
    val: T

    def __init__(self, val: T):
        super().__init__()
        self.val = val

    def unify_internal(self, other: "BaseType", mode: UnifyMode):
        if type(self) != type(other) or self.val != other.val:
            from transpiler.phases.typing.exceptions import TypeMismatchError, TypeMismatchKind
            raise TypeMismatchError(self, other, TypeMismatchKind.DIFFERENT_TYPE)

    def contains_internal(self, other: "BaseType") -> bool:
        return False

    def __str__(self):
        return str(self.val)

    def clone(self) -> "BaseType":
        return type(self)(self.val)

class BuiltinFeature(MagicType):
    pass


cur_var = 0

def next_var_id():
    global cur_var
    cur_var += 1
    return cur_var


@dataclass(eq=False)
class TypeVariable(BaseType):
    name: str = field(default_factory=lambda: next_var_id())
    resolved: Optional[BaseType] = None
    patch_attrs: dict = field(default_factory=dict)

    def __setattr__(self, key, value):
        if "patch_attrs" in self.__dict__ and key not in self.__dict__:
            self.patch_attrs[key] = value
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item):
        if "patch_attrs" in self.__dict__ and item in self.patch_attrs:
            return self.patch_attrs[item]
        raise AttributeError(item)

    def __str__(self):
        if self.resolved is None:
            #return f"TypeVar[\"{self.name}\"]"
            return f"_{self.name}"
        return str(self.resolved)

    def resolve(self) -> BaseType:
        if self.resolved is None:
            return self
        return self.resolved.resolve()

    def unify_internal(self, other: BaseType, mode: UnifyMode):
        if self is not other:
            if other.contains(self):
                from transpiler.phases.typing.exceptions import RecursiveTypeUnificationError
                raise RecursiveTypeUnificationError(self, other)
            self.resolved = other
            for k, v in self.patch_attrs.items():
                setattr(other, k, v)

    def contains_internal(self, other: BaseType) -> bool:
        return self.resolve() is other.resolve()

    def gen_sub(self, this: "BaseType", typevars, cache=None) -> "Self":
        if match := typevars.get(self.name):
            return match
        return self


GenMethodFactory = Callable[["BaseType"], "FunctionType"]


@dataclass(eq=False)
class TypeOperator(BaseType, ABC):
    args: List[BaseType]
    name: str = None
    variadic: bool = False
    optional_at: Optional[int] = None
    gen_methods: ClassVar[Dict[str, GenMethodFactory]] = {}
    gen_parents: ClassVar[List[BaseType]] = []
    is_protocol: bool = False
    is_protocol_gen: ClassVar[bool] = False
    match_cache: set["TypeOperator"] = field(default_factory=set, init=False)
    is_reference: bool = False
    is_intermediary: bool = False

    @staticmethod
    def make_type(name: str):
        class BuiltinType(TypeOperator):
            def __init__(self):
                super().__init__([], name)
        return BuiltinType()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.gen_methods = {}
        cls.gen_parents = []

    def __post_init__(self):
        assert all(x is not None for x in self.args)
        if self.name is None:
            self.name = self.__class__.__name__
        for name, factory in self.gen_methods.items():
            self.fields[name] = MemberDef(factory(self), ())
        for gp in self.gen_parents:
            if not isinstance(gp, BaseType):
                gp = gp(self.args)
            self.parents.append(gp)
            self.fields = {**gp.fields, **self.fields}
        self.is_protocol = self.is_protocol or self.is_protocol_gen
        self._add_default_eq()

    def _add_default_eq(self):
        if "__eq__" not in self.fields:
            if "DEFAULT_EQ" in globals():
                self.fields["__eq__"] = MemberDef(DEFAULT_EQ)

    def matches_protocol(self, protocol: "TypeOperator"):
        if hash(protocol) in self.match_cache:
            return
        from transpiler.phases.typing.exceptions import ProtocolMismatchError, TypeMismatchError
        try:
            dupl = protocol.gen_sub(self, {v.name: (TypeVariable(v.name) if isinstance(v.resolve(), TypeVariable) else v) for v in protocol.args})
            self.match_cache.add(hash(protocol))
            for name, ty in dupl.fields.items():
                if name == "__eq__":
                    continue
                if name not in self.fields:
                    raise ProtocolMismatchError(self, protocol, f"missing method {name}")
                corresp = self.fields[name].type
                corresp.remove_self().unify(ty.type.remove_self())
        except TypeMismatchError as e:
            if hash(protocol) in self.match_cache:
                self.match_cache.remove(hash(protocol))
            raise ProtocolMismatchError(self, protocol, e)

    def unify_internal(self, other: BaseType, mode: UnifyMode):
        from transpiler.phases.typing.exceptions import TypeMismatchError, TypeMismatchKind
        # TODO(zdimension): this is really broken... but it would be nice
        # if from_node := next(filter(None, (getattr(x, "from_node", None) for x in (other, self))), None):
        #     TB_NODE = from_node
        if not isinstance(other, TypeOperator):
            raise TypeMismatchError(self, other, TypeMismatchKind.DIFFERENT_TYPE)
        if mode.match_protocol:
            if other.is_protocol and not self.is_protocol:
                return other.unify_internal(self, mode)
            if self.is_protocol and not other.is_protocol:
                return other.matches_protocol(self)  # TODO: doesn't print the correct type in the error message
        assert self.is_protocol == other.is_protocol
        if type(self) != type(other): # and ((TY_NONE not in {self, other}) or isinstance(({self, other} - {TY_NONE}).pop(), UnionType)):
            if mode.search_hierarchy:
                if self.inherits_from(other) or other.inherits_from(self):
                    return
            # for parent in other.get_parents():
            #     try:
            #         self.unify(parent)
            #     except TypeMismatchError:
            #         pass
            #     else:
            #         return
            # for parent in self.get_parents():
            #     try:
            #         parent.unify(other)
            #     except TypeMismatchError:
            #         pass
            #     else:
            #         return
            raise TypeMismatchError(self, other, TypeMismatchKind.DIFFERENT_TYPE)
        if len(self.args) < len(other.args):
            return other.unify_internal(self, mode)
        if True or len(self.args) == 0: # todo: why check len?
            if self.name != other.name:
                raise TypeMismatchError(self, other, TypeMismatchKind.DIFFERENT_TYPE)
        for i, (a, b) in enumerate(zip_longest(self.args, other.args)):
            if a is None and self.variadic or b is None and other.variadic:
                continue

            if a is not None and b is None:
                if self.optional_at is not None and i >= self.optional_at:
                    continue
                else:
                    if getattr(other, "is_python_func", False):
                        other.args.append(a)
                        continue
                    else:
                        from transpiler.phases.typing.exceptions import ArgumentCountMismatchError
                        raise ArgumentCountMismatchError(*sorted((self, other), key=lambda x: x.is_intermediary))

            if isinstance(a, BaseType) and isinstance(b, BaseType):
                a.unify(b)
            else:
                if a != b:
                    raise TypeMismatchError(a, b, TypeMismatchKind.DIFFERENT_TYPE)

    def contains_internal(self, other: "BaseType") -> bool:
        return any(arg.contains(other) for arg in self.args)

    def __str__(self):
        return self.name + (f"<{', '.join(map(str, self.args))}>" if self.args else "")

    def __hash__(self):
        return hash((self.name, tuple(self.args)))

    def gen_sub(self, this: BaseType, typevars, cache=None) -> "Self":
        cache = cache or {}
        if me := cache.get(self):
            return me
        if len(self.args) == 0:
            return self
        assert all(x is not None for x in self.args)
        res = object.__new__(self.__class__)  # todo: ugly... should make a clone()
        cache[self] = res
        if isinstance(this, TypeOperator) and not isinstance(this, FunctionType):
            vardict = dict(zip(typevars.keys(), this.args))
        else:
            vardict = typevars
        for k, v in self.__dict__.items():
            setattr(res, k, v)
        res.args = [arg.resolve().gen_sub(this, vardict, cache) for arg in self.args]
        res.fields = {k: dataclasses.replace(v, type=v.type.gen_sub(this, vardict, cache)) for k, v in self.fields.items()}
        res.parents = [p.gen_sub(this, vardict, cache) for p in self.parents]
        #res.is_protocol = self.is_protocol
        return res

    def to_list(self) -> List["BaseType"]:
        return [self, *self.args]

@dataclass
class ModuleType(TypeOperator):
    is_python: bool = False


class FunctionType(TypeOperator):
    is_python_func: bool = False
    python_func_used: bool = False
    is_method: bool = False

    def __iter__(self):
        x = 5
        pass
        return iter([str(self)])

    def __init__(self, args: List[BaseType], ret: BaseType):
        super().__init__([ret, *args])

    @property
    def parameters(self):
        return self.args[1:]

    @property
    def return_type(self):
        return self.args[0]

    @return_type.setter
    def return_type(self, value):
        self.args[0] = value

    def __str__(self):
        ret, *args = map(str, self.args)
        if self.optional_at is not None:
            args = args[:self.optional_at] + [f"{x}=..." for x in args[self.optional_at:]]
        if self.variadic:
            args.append("*args")
        if args:
            args = f"{', '.join(args)}"
        else:
            args = ""
        return f"({args}) -> {ret}"

    def remove_self(self):
        res = FunctionType(self.parameters[1:], self.return_type)
        res.is_method = self.is_method
        res.variadic = self.variadic
        res.optional_at = self.optional_at - 1 if self.optional_at is not None else None
        return res


class CppType(TypeOperator):
    def __init__(self, name: str):
        super().__init__([name], name)

    def __str__(self):
        return self.name


class Union(TypeOperator):
    def __init__(self, left: BaseType, right: BaseType):
        super().__init__([left, right], "Union")


class TypeType(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "Type")

    @property
    def type_object(self) -> BaseType:
        return self.args[0]

    @type_object.setter
    def type_object(self, value: BaseType):
        self.args[0] = value


TY_OBJECT = TypeOperator.make_type("object")
TY_SELF = TypeOperator.make_type("Self")
def self_gen_sub(this, typevars, _):
    if this is not None:
        return this
    return TY_SELF
TY_SELF.gen_sub = self_gen_sub
TY_BOOL = TypeOperator.make_type("bool")
TY_TYPE = TypeOperator.make_type("type")
TY_INT = TypeOperator.make_type("int")
TY_FLOAT = TypeOperator.make_type("float")
TY_STR = TypeOperator.make_type("str")
TY_BYTES = TypeOperator.make_type("bytes")
TY_COMPLEX = TypeOperator.make_type("complex")
TY_NONE = TypeOperator.make_type("NoneType")
#TY_MODULE = TypeOperator([], "module")
TY_VARARG = TypeOperator.make_type("vararg")
TY_SLICE = TypeOperator.make_type("slice")


class PyList(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "list")

    @property
    def element_type(self):
        return self.args[0]


class PySet(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "set")

    @property
    def element_type(self):
        return self.args[0]


class PyDict(TypeOperator):
    def __init__(self, key: BaseType, value: BaseType):
        super().__init__([key, value], "dict")

    @property
    def key_type(self):
        return self.args[0]

    @property
    def value_type(self):
        return self.args[1]


class PyIterator(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "iter")

    @property
    def element_type(self):
        return self.args[0]


class TupleType(TypeOperator):
    def __init__(self, *args: List[BaseType]):
        super().__init__(args, "tuple")


class PromiseKind(Enum):
    TASK = 0
    JOIN = 1
    FUTURE = 2
    FORKED = 3
    GENERATOR = 4


class Promise(TypeOperator, ABC):
    def __init__(self, ret: BaseType, kind: PromiseKind):
        super().__init__([ret, MagicType(kind)])

    @property
    def return_type(self) -> BaseType:
        return self.args[0]

    @property
    def kind(self) -> PromiseKind:
        return self.args[1].val

    @kind.setter
    def kind(self, value: PromiseKind):
        if value == PromiseKind.GENERATOR:
            f_iter = FunctionType([], self)
            f_iter.is_method = True
            self.fields["__iter__"] = MemberDef(f_iter, ())
            f_next = FunctionType([], self.return_type)
            f_next.is_method = True
            self.fields["__next__"] = MemberDef(f_next, ())
        self.args[1].val = value

    def __str__(self):
        return f"{self.kind.name.lower()}<{self.return_type}>"

    def get_parents(self) -> List["BaseType"]:
        if self.kind == PromiseKind.GENERATOR:
            return [*super().get_parents()]
        return super().get_parents()


class Forked(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.FORKED)


class Task(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.TASK)


class Future(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.FUTURE)


class UserType(TypeOperator):
    def __init__(self, name: str):
        super().__init__([], name=name, is_reference=True)

    def unify_internal(self, other: "BaseType", mode: UnifyMode):
        if type(self) != type(other):
            from transpiler.phases.typing.exceptions import TypeMismatchError, TypeMismatchKind
            raise TypeMismatchError(self, other, TypeMismatchKind.DIFFERENT_TYPE)

class UnionType(TypeOperator):
    def __init__(self, *args: List[BaseType]):
        super().__init__(args, "Union")
        self.parents.extend(set(args))

    def is_optional(self):
        if len(self.args) == 2 and TY_NONE in self.args:
            return (set(self.args) - {TY_NONE}).pop()
        return False

